include(defines.m4)dnl
changequote(`[!',`!]')dnl
dnl define([!M4_HANDLE_EOT!],[!1!])dnl
dnl define([!M4_HANDLE_PRODUCER!],[!1!])dnl
@startuml
actor aUser
note left of aUser: M4_LAST_GENERATED

participant anInputPlugin
participant "<i>YASAPI/NT</i>\n(thePlayer)" as thePlayer
participant "thePlayer.engine\ndriven by <i>Winamp</i> threads\n(maintains <i>eState</i>)" as theEngine
participant "thePlayer.engine\ndriven by a background thread\n(maintains <i>eState</i>)" as theThread
participant "thePlayer.engine.cs\n(<i>CriticalSection</i>)" as theCriticalSection
ifdef([!M4_HANDLE_EOT!],
[!participant "thePlayer.engine.thread.hEot\n(event <i>End of Track</i>)" as theEventEndOfTrack
!],
[!ifdef([!M4_HANDLE_PRODUCER!],
[!participant "thePlayer.engine.thread.hProducer" as theEventProducer
!])dnl
participant "thePlayer.engine.thread.hConsumer" as theEventConsumer
!])dnl
participant "thePlayer.engine.pAudioClient\n(<i>WASAPI</i>)" as theAudioClient
participant "thePlayer.engine.pAudioRenderClient\n(<i>WASAPI</i>)" as theAudioRenderClient
participant "<i>YASAPI/NT</i>'s buffer\n(thePlayer.engine.ring)" as theRing

'[
aUser --> anInputPlugin : play()
activate anInputPlugin

'[
loop
'[
anInputPlugin --> thePlayer : PlayerOpen()
activate thePlayer
'[
thePlayer --> theEngine : EngineCreate()
activate theEngine
'[
alt <i>eState < eEngineBase</i>
theEngine --> theEngine : EngineCreateBase()
activate theEngine
theEngine --> theCriticalSection : InitializeCriticalSection()
note right: <i>Windows</i> API
ifdef([!M4_HANDLE_EOT!],
[!theEngine --> theEventEndOfTrack : CreateEvent()
note right: <i>Windows</i> API
!],
[!ifdef([!M4_HANDLE_PRODUCER!],
theEngine --> theEventProducer : CreateEvent()
!])
note right: <i>Windows</i> API
theEngine --> theEventConsumer : CreateEvent()
note right: <i>Windows</i> API
!])dnl
theEngine --> theThread : CreateThread()
note right: <i>Windows</i> API
theEngine --> theEngine : eEngineNull → eEngineBase
deactivate theEngine
end
']

theEngine --> theCriticalSection : EnterCriticalSection()
note right: <i>Windows</i> API
activate theCriticalSection
theCriticalSection --> theEngine
deactivate theCriticalSection
'[
alt eState < eEngineInitialized or the input format has changed
'[
loop while eEngineBase < eState

'[
ifdef([!M4_HANDLE_EOT!],
[!theEngine --> theEventEndOfTrack : ResetEvent()
note right: <i>Windows</i> API
activate theEventEndOfTrack
theEventEndOfTrack --> theEngine
deactivate theEventEndOfTrack
!],
[!theEngine --> theEventConsumer : ResetEvent()
note right: <i>Windows</i> API
activate theEventConsumer
theEventConsumer --> theEngine
deactivate theEventConsumer
!])dnl
']

theEngine --> theCriticalSection : LeaveCriticalSection()
note right: <i>Windows</i> API
activate theCriticalSection
theCriticalSection --> theEngine
deactivate theCriticalSection

'[
ifdef([!M4_HANDLE_EOT!],
[!theEngine --> theEventEndOfTrack : WaitForSingleObject()
note right: <i>Windows</i> API: We're just waiting for the background thread to finish\nconsuming <i>YASAPI/NT</i>'s buffer and then immediatly continue.
activate theEventEndOfTrack
theEventEndOfTrack --> theEngine
deactivate theEventEndOfTrack
!],
[!theEngine --> theEventConsumer : WaitForSingleObject()
note right: <i>Windows</i> API: We're just waiting for the background thread to finish\nconsuming <i>YASAPI/NT</i>'s buffer and then immediatly continue.
activate theEventConsumer
theEventConsumer --> theEngine
deactivate theEventConsumer
!])dnl
']

theEngine --> theCriticalSection : EnterCriticalSection()
note right: <i>Windows</i> API
activate theCriticalSection
theCriticalSection --> theEngine
deactivate theCriticalSection

end
']
theEngine --> theEngine : EngineCreateGaplessSession()
'note right: Note that's not sufficient: The background thread may have finished consuming\n<i>YASAPI/NT</i>'s buffer before we'e going to write for the first time.
activate theEngine
theEngine --> theAudioClient : IMMDevice::Activate()
note right: <i>WASAPI</i>
theEngine --> theAudioClient : IAudioClient::Initialize()
note right: <i>WASAPI</i>
activate theAudioClient
theAudioClient --> theEngine
deactivate theAudioClient
theEngine --> theEngine : eEngineBase → eEngineInitialized
deactivate theEngine
end
']
theEngine --> theCriticalSection : LeaveCriticalSection()
note right: <i>Windows</i> API
activate theCriticalSection
theCriticalSection --> theEngine
deactivate theCriticalSection


theEngine --> thePlayer
deactivate theEngine
']
thePlayer --> anInputPlugin
deactivate thePlayer
']

anInputPlugin --> aUser
deactivate anInputPlugin
']

loop
anInputPlugin --> thePlayer : PlayerWrite()
'[
activate thePlayer
thePlayer --> theEngine : EngineWrite()
'[
activate theEngine

theEngine --> theCriticalSection : EnterCriticalSection()
note right: <i>Windows</i> API
activate theCriticalSection
theCriticalSection --> theEngine
deactivate theCriticalSection

alt <i>ring</i> is not instantiated
theEngine --> theRing : RingCreateFrames()
note right: Instantiate <i>YASAPI/NT</i>'s\nbuffer before writing to it!
end
theEngine --\ theRing : RingWrite()
note right: asynchronously
'[
activate theRing
theRing --> theEngine
deactivate theRing
']

alt <i>eState < eEngineStarted</i> and the minimum amount of frames is written to <i>ring</i>
theEngine --> theThread : EngineSend (eCommandStart)

'[
activate theThread
'[
theThread --> theThread : EngineConnect()
activate theThread

theThread --> theAudioClient : IAudioClient::GetService()
note right: <i>WASAPI</i>
'[
activate theAudioClient
theAudioClient --> theAudioRenderClient
theAudioClient --> theThread
deactivate theAudioClient
']

'[
theThread --> theAudioClient : IAudioClient::Start()
activate theAudioClient
note right: <i>WASAPI</i>
theAudioClient --> theThread
deactivate theAudioClient
']

theThread --> theThread
deactivate theThread
']
theThread --> theEngine
deactivate theThread



'[
loop in the background periodically triggered by <i>timer.hEvent</i> or <i>timer.hTimer</i> when in <i>exclusive</i> or <i>shared</i> mode, respectively
'[
theThread --> theThread : ThreadPlay()
activate theThread

theThread --> theCriticalSection : EnterCriticalSection()
note right: <i>Windows</i> API
activate theCriticalSection
theCriticalSection --> theThread
deactivate theCriticalSection
'[
alt <i>eState < eEngineDraining</i> or <i>ring</i> is not empty

theThread --> theThread : EngineRender
activate theThread

theThread --> theAudioRenderClient : IAudioRenderClient::GetBuffer()
note right: <i>WASAPI</i>
activate theAudioRenderClient
theAudioRenderClient --> theThread
deactivate theAudioRenderClient

theThread --/ theRing : RingReadFrames()
note right: asynchronously
activate theRing
theRing --> theThread
deactivate theRing

theThread --> theAudioRenderClient : IAudioRenderClient::ReleaseBuffer()
note right: <i>WASAPI</i>
activate theAudioRenderClient
theAudioRenderClient --> theThread
deactivate theAudioRenderClient

deactivate theRing

theThread --> theThread
deactivate theThread

'] [
else

theThread --> theThread : EngineKill(eEngineBase)
note right: state transition down to the argument state
activate theThread

'[
theThread --> theAudioClient : IAudioClient::Stop()
activate theAudioClient
note right: <i>WASAPI</i>
theAudioClient --> theThread
deactivate theAudioClient
']

theThread --> theAudioRenderClient : IUnknown::Release()
note right: <i>WASAPI</i>
activate theAudioRenderClient
theAudioRenderClient --> theThread
deactivate theAudioRenderClient

theThread --> theRing : RingDestroy()
activate theRing
theRing --> theThread
deactivate theRing

theThread --> theAudioClient : IUnknown::Release()
note right: <i>WASAPI</i>
activate theAudioClient
theAudioClient --> theThread
deactivate theAudioClient

theThread --> theThread : eEngineDraining → eEngineBase

'[
ifdef([!M4_HANDLE_EOT!],
[!theThread --> theEventEndOfTrack : SetEvent()
note right: <i>Windows</i> API: Notify the foreground thread in case it's waiting.
activate theEventEndOfTrack
theEventEndOfTrack --> theThread
deactivate theEventEndOfTrack
!],
[!theThread --> theEventConsumer : SetEvent()
note right: <i>Windows</i> API: Notify the foreground thread in case it's waiting.
activate theEventConsumer
theEventConsumer --> theThread
deactivate theEventConsumer
!])dnl
']

theThread --> theThread
deactivate theThread

end
']
theThread --> theCriticalSection : LeaveCriticalSection()
note right: <i>Windows</i> API
activate theCriticalSection
theCriticalSection --> theThread
deactivate theCriticalSection

theThread --> theThread

deactivate theThread


']
end
']

']
theEngine --> theEngine : eEngineInitialized → eEngineStarted
end

theEngine --> theCriticalSection : LeaveCriticalSection()
note right: <i>Windows</i> API
activate theCriticalSection
theCriticalSection --> theEngine
deactivate theCriticalSection

theEngine --> thePlayer
deactivate theEngine
']
thePlayer --> anInputPlugin
deactivate thePlayer
']
end

'[
loop while <i>PlayerIsPlaying()</i> is "yes"
anInputPlugin --> thePlayer : PlayerIsPlaying()
activate thePlayer

thePlayer --> theEngine : EngineIsPlaying()
activate theEngine

'[
theEngine --> theCriticalSection : EnterCriticalSection()
note right: <i>Windows</i> API
activate theCriticalSection
theCriticalSection --> theEngine
deactivate theCriticalSection
']

theEngine --> theEngine : eEngineStarted → eEngineDraining

'[
theEngine --> theCriticalSection : LeaveCriticalSection()
note right: <i>Windows</i> API
activate theCriticalSection
theCriticalSection --> theEngine
deactivate theCriticalSection
']

theEngine --> thePlayer : "no"
note right: Of course we're sheeting because the background thread is still runing and consumes what's\ncached in <i>YASAPI/NT</i>'s buffer. That's exactly what we want for gapless playback: If possible\nwith the next track we'd like to simply continue writing to <i>YASAPI/NT</i>'s buffer!

deactivate theEngine

thePlayer --> anInputPlugin : "no"
deactivate thePlayer
end
']

'[
anInputPlugin --> thePlayer : PlayerClose()
note right: In gapless mode there's nothing left to do.
activate thePlayer
thePlayer --> anInputPlugin
deactivate thePlayer
']

end
']

@enduml
