/*
 * yasapi_engine_client_state.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
static int EngineClientStateCacheEnter(EngineClient *pClient)
{
  if (FsmUp(pClient->pPlayer,ePlayerStateCache,__FILE__,__LINE__,__func__)<0)
    goto exit;

  ++pClient->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineClientStateCacheLeave(EngineClient *pClient)
{
DMARKLN();
  --pClient->fsm.pState;
DMARKLN();
}

///////////////////////////////////////////////////////////////////////////////
static int EngineClientStateTimerEventEnter(EngineClient *pClient)
{
  if (!(pClient->timer.hEvent=CreateEventW(NULL,FALSE,FALSE,NULL))) {
    DMESSAGE("creating timer event");
    goto exit;
  }

  ++pClient->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineClientStateTimerEventLeave(EngineClient *pClient)
{
DMARKLN();
  --pClient->fsm.pState;
  CloseHandle(pClient->timer.hEvent);
DMARKLN();
}

///////////////////////////////////////////////////////////////////////////////
static int EngineClientStateTimerEnter(EngineClient *pClient)
{
  if (!(pClient->timer.hTimer=CreateWaitableTimerW(NULL,FALSE,NULL))) {
    DMESSAGE("creating timer event");
    goto exit;
  }

  ++pClient->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineClientStateTimerLeave(EngineClient *pClient)
{
DMARKLN();
  --pClient->fsm.pState;
  CloseHandle(pClient->timer.hTimer);
DMARKLN();
}

///////////////////////////////////////////////////////////////////////////////
static int EngineClientStateThreadEventEnter(EngineClient *pClient)
{
  // This is a MANUAL RESET event!
  if (!(pClient->thread.hEvent=CreateEventW(NULL,TRUE,FALSE,NULL))) {
    DMESSAGE("creating consumer event");
    goto exit;
  }

  ++pClient->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineClientStateThreadEventLeave(EngineClient *pClient)
{
DMARKLN();
  --pClient->fsm.pState;
  CloseHandle(pClient->thread.hEvent);
DMARKLN();
}

///////////////////////////////////////////////////////////////////////////////
static int EngineClientStateThreadEnter(EngineClient *pClient)
{
  pClient->thread.hThread=CreateThread(
    NULL,       // _In_opt_  LPSECURITY_ATTRIBUTES  lpThreadAttributes,
    0ul,        // _In_      SIZE_T                 dwStackSize,
    EngineServerThreadProc,
                // _In_      LPTHREAD_START_ROUTINE lpStartAddress,
    pClient,  // _In_opt_  LPVOID                 lpParameter,
    0ul,        // _In_      DWORD                  dwCreationFlags,
                // _Out_opt_ LPDWORD                lpThreadId
    &pClient->thread.dwThreadId
  );

  if (INVALID_HANDLE_VALUE==pClient->thread.hThread) {
    DMESSAGE("starting consumer thread");
    goto exit;
  }

  ++pClient->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineClientStateThreadLeave(EngineClient *pClient)
{
DMARKLN();
  --pClient->fsm.pState;
  WaitForSingleObject(pClient->thread.hThread,INFINITE);
  CloseHandle(pClient->thread.hThread);
  pClient->thread.hThread=INVALID_HANDLE_VALUE;
DMARKLN();
}

///////////////////////////////////////////////////////////////////////////////
static int EngineClientStateThreadRunningEnter(EngineClient *pClient)
{
  int code=-1;
  IFsmState *pState=pClient->fsm.pStateMin+eEngineClientStateThreadRunning;
  Player *pPlayer=pClient->pPlayer;
  int nLockRecursion;
  int dwWait;

  PlayerLock(pPlayer ); // {

  while (pClient->fsm.pState<pState) {
    ResetEvent(pClient->thread.hEvent);
    PlayerUnlockDeep(pPlayer,&nLockRecursion); // {
    dwWait=WaitForSingleObject(pClient->thread.hEvent,INFINITE);
    PlayerLockDeep(pPlayer,nLockRecursion ); // }

    switch (dwWait) {
    case WAIT_ABANDONED:
      DVMESSAGE("WAIT_ABANDONED %s",FsmStateId(pClient));
      goto exit;
    case WAIT_TIMEOUT:
      DVMESSAGE("WAIT_TIMEOUT %s",FsmStateId(pClient));
      goto exit;
    case WAIT_FAILED:
      DVMESSAGE("WAIT_FAILED %s: %lu",FsmStateId(pClient),GetLastError());
      goto exit;
    default:
      break;
    }
  }

  code=0;
exit:
  PlayerUnlock(pPlayer); // }
  DWRITELN("!!! THREAD RUNNING !!!!!!!!!!!!!!!!!!!!!!");

  return code;
}

static void EngineClientStateThreadRunningLeave(EngineClient *pClient)
{
  IFsmState *pState=pClient->fsm.pStateMin+eEngineClientStateThreadRunning-1;
  EngineServer *pServer=pClient->pServer;
  Player *pPlayer=pClient->pPlayer;
  int nLockRecursion;
  int dwWait;

DMARKLN();
  PlayerLock(pPlayer ); // {
  pServer->thread.eStateRequest=eEngineServerStateLoopEnd;
  SetEvent(pServer->thread.hEvent);

  while (pState<pClient->fsm.pState) {
    ResetEvent(pClient->thread.hEvent);
    PlayerUnlockDeep(pPlayer,&nLockRecursion); // {
    dwWait=WaitForSingleObject(pClient->thread.hEvent,INFINITE);
    PlayerLockDeep(pPlayer,nLockRecursion ); // }

    switch (dwWait) {
    case WAIT_ABANDONED:
      DVMESSAGE("WAIT_ABANDONED %s",FsmStateId(pClient));
      goto exit;
    case WAIT_TIMEOUT:
      DVMESSAGE("WAIT_TIMEOUT %s",FsmStateId(pClient));
      goto exit;
    case WAIT_FAILED:
      DVMESSAGE("WAIT_FAILED %s: %lu",FsmStateId(pClient),GetLastError());
      goto exit;
    default:
      break;
    }
  }
exit:
  PlayerUnlock(pPlayer); // }
DMARKLN();
}

///////////////////////////////////////////////////////////////////////////////
static int EngineClientStateCloseEnter(EngineClient *pClient)
{
  ++pClient->fsm.pState;

  return 0;
}

static void EngineClientStateCloseLeave(EngineClient *pClient)
{
DMARKLN();
  --pClient->fsm.pState;
DMARKLN();
}


///////////////////////////////////////////////////////////////////////////////
static int EngineClientStateAudioClientEnter(EngineClient *pClient)
{
  IMMDevice *pDevice=pClient->param.pDevice;
  IAudioClient *pAudioClient;
  HRESULT hr;

  hr=pDevice->lpVtbl->Activate(pDevice,
    &IID_IAudioClient,      // [in]  REFIID      iid,
    CLSCTX_ALL,             // [in]  DWORD       dwClsCtx,
    NULL,                   // [in]  PROPVARIANT *pActivationParams,
    (void **)&pAudioClient  // [out] void        **ppInterface
  );

  if (FAILED(hr)) {
#define DCLABEL "IMMDevice::Activate<IAudioClient>" // [
    switch (hr) {
    DCMESSAGE(DCLABEL,E_NOINTERFACE);
    DCMESSAGE(DCLABEL,E_POINTER);
    DCMESSAGE(DCLABEL,E_INVALIDARG);
    DCMESSAGE(DCLABEL,E_OUTOFMEMORY);
    DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_INVALIDATED);
    DDMESSAGE(DCLABEL,hr);
    }
#undef DCLABEL // ]
    goto exit;
  }

  pClient->pAudioClient=pAudioClient;
  pAudioClient->lpVtbl->Reset(pAudioClient);
  ++pClient->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineClientStateAudioClientLeave(EngineClient *pClient)
{
  IAudioClient **ppAudioClient=&pClient->pAudioClient;

DMARKLN();
  --pClient->fsm.pState;

  if (*ppAudioClient) {
    (*ppAudioClient)->lpVtbl->Release(*ppAudioClient);
    *ppAudioClient=NULL;
  }
DMARKLN();
}

///////////////////////////////////////////////////////////////////////////////
static int EngineClientStateDevicePeriodEnter(EngineClient *pClient)
{
  IAudioClient *pAudioClient=pClient->pAudioClient;
  REFERENCE_TIME hnsDevicePeriod;
  HRESULT hr;

  /////////////////////////////////////////////////////////////////////////////
  if (0ll<pClient->param.hnsDevicePeriod)
    goto success;

  /////////////////////////////////////////////////////////////////////////////
  if (pClient->options.device.eMode) {
    hr=pAudioClient->lpVtbl->GetDevicePeriod(pAudioClient,
      NULL,       // [out] REFERENCE_TIME *phnsDefaultDevicePeriod,
      &hnsDevicePeriod
                  // [out] REFERENCE_TIME *phnsMinimumDevicePeriod
    );
  }
  else {
    hr=pAudioClient->lpVtbl->GetDevicePeriod(pAudioClient,
      &hnsDevicePeriod,
                  // [out] REFERENCE_TIME *phnsDefaultDevicePeriod,
      NULL        // [out] REFERENCE_TIME *phnsMinimumDevicePeriod
    );
  }

  if (FAILED(hr)) {
#define DCLABEL "IAudioClient::GetDevicePeriod" // [
    switch (hr) {
    DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_SERVICE_NOT_RUNNING);
    DCMESSAGE(DCLABEL,E_POINTER);
    DDMESSAGE(DCLABEL,hr);
    }
#undef DCLABEL // ]
    goto exit;
  }
success:
  DVWRITELN("*** hnsDevicePeriod=%I64d ***",hnsDevicePeriod);
  pClient->param.hnsDevicePeriod=hnsDevicePeriod;
  ++pClient->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineClientStateDevicePeriodLeave(EngineClient *pClient)
{
DMARKLN();
  --pClient->fsm.pState;
DMARKLN();
}

///////////////////////////////////////////////////////////////////////////////
static int EngineClientStateBufferSizeEnter(EngineClient *pClient)
{
  IAudioClient *pAudioClient=pClient->pAudioClient;
  UINT32 uNumFramesDevice;
  HRESULT hr;

  hr=pAudioClient->lpVtbl->GetBufferSize(pAudioClient,
    &uNumFramesDevice // UINT32 *pNumBufferFrames
  );

  if (FAILED(hr)) {
#define DCLABEL "IAudioClient::GetBufferSize" // [
    switch (hr) {
    DCMESSAGE(DCLABEL,AUDCLNT_E_NOT_INITIALIZED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_SERVICE_NOT_RUNNING);
    DCMESSAGE(DCLABEL,E_POINTER);
    DDMESSAGE(DCLABEL,hr);
    }
#undef DCLABEL // ]
    goto exit;
  }

  pClient->param.uNumFramesDevice=uNumFramesDevice;

  if (eStrategyPush==pClient->options.device.eStrategy) {
    pClient->param.DueTime.QuadPart
        =-0.5*YASAPI_FRAMES2HNS(uNumFramesDevice,&pClient->wfx.target)-0.5;

    if (pClient->options.device.bTimerPeriodic) {
      pClient->param.lPeriod
          =0.5*YASAPI_FRAMES2MS(uNumFramesDevice,&pClient->wfx.target)+0.5;
    }
    else
      pClient->param.lPeriod=0l;
  }
  else {
    pClient->param.DueTime.QuadPart=0ll;
    pClient->param.lPeriod=0l;
  }

  ++pClient->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineClientStateBufferSizeLeave(EngineClient *pClient)
{
DMARKLN();
  --pClient->fsm.pState;
DMARKLN();
}

///////////////////////////////////////////////////////////////////////////////
static int EngineClientStateAudioClientInitializeEnter(EngineClient *pClient)
{
  IAudioClient *pAudioClient=pClient->pAudioClient;
  HRESULT hr;

  /////////////////////////////////////////////////////////////////////////////
  if (eStrategyPull==pClient->options.device.eStrategy) {
    hr=pAudioClient->lpVtbl->Initialize(pAudioClient,
      pClient->param.eShareMode,
                            // [in]       AUDCLNT_SHAREMODE ShareMode,
      pClient->param.dwStreamFlags,
                            // [in]       DWORD             StreamFlags,
      pClient->param.hnsDevicePeriod,
                            // [in]       REFERENCE_TIME    hnsBufferDuration,
      pClient->param.hnsDevicePeriod,
                            // [in]       REFERENCE_TIME    hnsPeriodicity,
      &pClient->wfx.target.Format,
                            // [in] const WAVEFORMATEX      *pFormat,
      NULL                  // [in]       LPCGUID           AudioSessionGuid
    );
  }
  else {
    hr=pAudioClient->lpVtbl->Initialize(pAudioClient,
      pClient->param.eShareMode,
                            // [in]       AUDCLNT_SHAREMODE ShareMode,
      pClient->param.dwStreamFlags,
                            // [in]       DWORD             StreamFlags,
      pClient->param.hnsDevicePeriod,
                            // [in]       REFERENCE_TIME    hnsBufferDuration,
      0ull,                 // [in]       REFERENCE_TIME    hnsPeriodicity,
      &pClient->wfx.target.Format,
                            // [in] const WAVEFORMATEX      *pFormat,
      NULL                  // [in]       LPCGUID           AudioSessionGuid
    );
  }

  if (FAILED(hr)) {
#define DCLABEL "IAudioClient::Initialize" // [
    switch (hr) {
    DCMESSAGE(DCLABEL,CO_E_NOTINITIALIZED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_ALREADY_INITIALIZED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_WRONG_ENDPOINT_TYPE);
    case AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED:
      if (!pClient->options.device.bAlignmentDance) {
        DMESSAGE(DCLABEL " failed with AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED");
        break;
      }

      if (0<pClient->param.nAlignment) {
        DMESSAGE(DCLABEL " failed repeatedly with AUDCLNT_E_BUFFER_SIZE_NOT_ALIGNED");
        break;
      }

      // https://blogs.msdn.microsoft.com/matthew_van_eerde/2009/04/03/sample-wasapi-exclusive-mode-event-driven-playback-app-including-the-hd-audio-alignment-dance/
      pClient->param.hnsDevicePeriod=
          10.0*1000                             // (hns / ms) *
          *1000                                 // (ms / s) *
          *pClient->param.uNumFramesDevice    // frames /
          /pClient->wfx.target.Format.nSamplesPerSec
                                                // (frames / s)
          +0.5                                  // rounding
          ;
      DVWRITELN("*** uNumFramesDevice (first pass)=%u ***",
          pClient->param.uNumFramesDevice);
      DVWRITELN("*** hnsDevicePeriod (first pass)=%I64d ***",
          pClient->param.hnsDevicePeriod);

      ++pClient->param.nAlignment;
      ++pClient->fsm.pState;
      return FsmBacktrack(pClient,eEngineClientStateAudioClient,__FILE__,__LINE__,__func__);
    DCMESSAGE(DCLABEL,AUDCLNT_E_BUFFER_SIZE_ERROR);
    DCMESSAGE(DCLABEL,AUDCLNT_E_CPUUSAGE_EXCEEDED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_IN_USE);
    DCMESSAGE(DCLABEL,AUDCLNT_E_ENDPOINT_CREATE_FAILED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_INVALID_DEVICE_PERIOD);
    DCMESSAGE(DCLABEL,AUDCLNT_E_EXCLUSIVE_MODE_NOT_ALLOWED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_BUFDURATION_PERIOD_NOT_EQUAL);
    DCMESSAGE(DCLABEL,AUDCLNT_E_SERVICE_NOT_RUNNING);
    DCMESSAGE(DCLABEL,E_POINTER);
    DCMESSAGE(DCLABEL,E_INVALIDARG);
    DCMESSAGE(DCLABEL,E_OUTOFMEMORY);
    DDMESSAGE(DCLABEL,hr);
    }
#undef DCLABEL // ]
    goto exit;
  }

  ++pClient->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineClientStateAudioClientInitializeLeave(EngineClient *pClient)
{
DMARKLN();
  --pClient->fsm.pState;
DMARKLN();
}

///////////////////////////////////////////////////////////////////////////////
static int EngineClientStateSetEventHandleEnter(EngineClient *pClient)
{
  IAudioClient *pAudioClient=pClient->pAudioClient;
  HRESULT hr;

  /////////////////////////////////////////////////////////////////////////////
  if (eStrategyPull==pClient->options.device.eStrategy) {
    hr=pAudioClient->lpVtbl->SetEventHandle(pAudioClient,
      pClient->timer.hEvent   // [in] HANDLE eventHandle
    );

    if (FAILED(hr)) {
#define DCLABEL "IAudioClient::SetEventHandle" // [
      switch (hr) {
      DCMESSAGE(DCLABEL,E_INVALIDARG);
      DCMESSAGE(DCLABEL,AUDCLNT_E_EVENTHANDLE_NOT_EXPECTED);
      DCMESSAGE(DCLABEL,AUDCLNT_E_NOT_INITIALIZED);
      DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_INVALIDATED);
      DCMESSAGE(DCLABEL,AUDCLNT_E_SERVICE_NOT_RUNNING);
      DDMESSAGE(DCLABEL,hr);
      }
#undef DCLABEL // ]
      goto exit;
    }
  }

  ++pClient->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineClientStateSetEventHandleLeave(EngineClient *pClient)
{
DMARKLN();
  --pClient->fsm.pState;
DMARKLN();
}

///////////////////////////////////////////////////////////////////////////////
static void EngineClientCeilFrames(UINT32 *puFrames, UINT32 uChunkFrames)
{
  UINT32 uMod=*puFrames%uChunkFrames;

  if (uMod)
    *puFrames+=uChunkFrames-uMod;
}

int EngineClientCreateRing(EngineClient *pClient, RingCreateFun *RingCreate,
    DWORD dwChunkSizeSource)
{
  GtkClient *pGtk=&pClient->pPlayer->gtk;
  Ring *pRing=&pClient->ring;
  ConfigWidget *pWidget=&pGtk->config.config.monitor.engine.num_frames_ring;
  ////////////////
  WAVEFORMATEX *pSource=&pClient->wfx.source.Format;
  UINT32 uFramesChunk=dwChunkSizeSource/pSource->nBlockAlign;
  double msLatency=pClient->options.device.xMaxLatencyAssumed;
  UINT32 uFramesRing=YASAPI_MS2FRAMES(msLatency,&pClient->wfx.target)+0.5;
  ////////////////
  double xPlaybackBegin=pClient->options.device.xPlaybackBegin;
  UINT32 uFramesDevice=pClient->param.uNumFramesDevice;
  UINT32 uFramesPlaybackBegin=xPlaybackBegin*uFramesDevice+0.5;

  /////////////////////////////////////////////////////////////////////////////
  DASSERT(dwChunkSizeSource,exit);

  /////////////////////////////////////////////////////////////////////////////
  pClient->param.uFramesPlaybackBegin=uFramesPlaybackBegin;

DVWRITELN("uFramesRing: %u ========================",uFramesRing);
  if (uFramesRing<uFramesPlaybackBegin)
    uFramesRing=uFramesPlaybackBegin;

DVWRITELN("uFramesRing: %u ========================",uFramesRing);
  /////////////////////////////////////////////////////////////////////////////
DVWRITELN("dwChunkSizeSource: %lu ========================",dwChunkSizeSource);
  EngineClientCeilFrames(&uFramesRing,uFramesChunk);
DVWRITELN("uFramesRing: %u (%u %u) =================",uFramesRing,
  uFramesRing/uFramesChunk,uFramesRing%uFramesChunk);

  pClient->param.uNumFramesRing=uFramesRing;
  GtkClientUpdateWidget(pGtk,pWidget);

  if (RingCreate(pRing,pClient,uFramesRing)<0) {
    DMESSAGE("creating ring");
    goto exit;
  }

  pClient->monitor.xStart=(double)uFramesPlaybackBegin/uFramesRing;
  GtkClientUpdateWidget(pGtk,&pGtk->config.config.monitor.start);

  return 0;
exit:
  return -1;
}

static int EngineClientStateRingEnter(EngineClient *pClient)
{
  if (EngineClientCreateRing(pClient,RingCreateFrames,pClient->param.dwChunkSizeSource)<0) {
    DMESSAGE("crating ring");
    goto exit;
  }

  ++pClient->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineClientStateRingLeave(EngineClient *pClient)
{
  GtkClient *pGtk=&pClient->pPlayer->gtk;

DMARKLN();
  --pClient->fsm.pState;
  RingDestroy(&pClient->ring);
  ZeroMemory(&pClient->param,sizeof pClient->param);
  EngineStatsReset(&pClient->stats);
  EngineMonitorReset(&pClient->monitor);
  GtkClientUpdateDialog(pGtk,&pGtk->config);
DMARKLN();
}

///////////////////////////////////////////////////////////////////////////////
static int EngineClientStateOpenEnter(EngineClient *pClient)
{

  if (EngineClientRequestServerState(pClient,eEngineServerStateOpen,__FILE__,__LINE__,__func__)<0) {
    DMESSAGE("requesting consumer to open");
    goto exit;
  }

  ++pClient->fsm.pState;

  return 0;
exit:
  return -1;
}

static void EngineClientStateOpenLeave(EngineClient *pClient)
{
DMARKLN();
  --pClient->fsm.pState;
  EngineClientRequestServerState(pClient,eEngineServerStateBase,__FILE__,__LINE__,__func__);
DMARKLN();
}

///////////////////////////////////////////////////////////////////////////////
IEngineClientState gcaIEngineClientState[eEngineClientStateCount]={
#define YASAPI_ENGINE_CLIENT(x) { \
  .pId=#x, \
  .Enter=EngineClientState ## x ## Enter, \
  .Leave=EngineClientState ## x ## Leave, \
},
#include <yasapi_engine_client_state.h>
#undef YASAPI_ENGINE_CLIENT
};
