/*
 * yasapi_server.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
static int EngineServerRender(EngineServer *pServer,
    EEngineServer eEngineServer, EEngineServer eEngineServerRequest);
static void EngineServerHandshake(EngineServer *pServer);
static void EngineServerError(EngineServer *pServer);

///////////////////////////////////////////////////////////////////////////////
DWORD WINAPI EngineServerThreadProc(LPVOID lpParameter)
{
  enum {
    eThreadEvent,
    eTimerEvent,
    eTimer,
    eCount,
  };

  EngineServer server;
  Player *pPlayer;
  HANDLE aHandles[eCount];
  int nLockRecursion;
  DWORD dwMilliseconds;
  DWORD dwWait;
  EEngineServer eEngineServer;
  EEngineServer eEngineServerRequest;

  if (EngineServerCreate(&server,lpParameter)<0)
    goto ecreate;

  if (FsmUp(&server,eEngineServerStateBase,__FILE__,__LINE__,__func__)<0)
    goto eopen;

  pPlayer=server.pClient->pPlayer;
  PlayerLock(pPlayer ); // {
  aHandles[eThreadEvent]=server.thread.hEvent;
  aHandles[eTimerEvent]=server.pClient->timer.hEvent;
  aHandles[eTimer]=server.pClient->timer.hTimer;
  ++server.pClient->fsm.pState;
  SetEvent(server.pClient->thread.hEvent);

  for (;;) {
    eEngineServer=FsmState(&server);

    switch (eEngineServer) {
    case eEngineServerStatePauseConnected:
    case eEngineServerStatePauseDisconnected:
      dwMilliseconds=INFINITE;
      break;
    default:
      if (eEngineServerStateOpen<=eEngineServer)
        dwMilliseconds=10000ul;
      else
        dwMilliseconds=INFINITE;

      break;
    }

    ResetEvent(server.thread.hEvent);
    PlayerUnlockDeep(pPlayer,&nLockRecursion); // {

    dwWait=WaitForMultipleObjects(
      eCount,         // _In_       DWORD  nCount,
      aHandles,       // _In_ const HANDLE *lpHandles,
      FALSE,          // _In_       BOOL   bWaitAll,
      dwMilliseconds  // _In_       DWORD  dwMilliseconds
    );

    PlayerLockDeep(pPlayer,nLockRecursion ); // }
    eEngineServer=FsmState(&server);
    eEngineServerRequest=server.thread.eStateRequest;

    switch (dwWait) {
    case WAIT_OBJECT_0+eThreadEvent:
      switch (eEngineServerRequest) {
      case eEngineServerStateLoopEnd:
        goto end;
      case eEngineServerStateOpen:
        if (eEngineServerRequest<=eEngineServer) {
          // Resetting the server is executed deferred with the next
          // playback cycle.
          break;
        }

        // Opening the server is executed immediatly.
        if (FsmUp(&server,eEngineServerRequest,__FILE__,__LINE__,__func__)<0) {
          DMESSAGE("opening server");
          EngineServerError(&server);
          goto end;
        }

        if (EngineServerRender(&server,eEngineServer,eEngineServerRequest)<0) {
          DMESSAGE("rendering");
          EngineServerError(&server);
          goto end;
        }

        EngineServerHandshake(&server);
        break;
      case eEngineServerStateKill:
        // Killing is executed deferred with the next playback cycle.
        break;
      case eEngineServerStatePauseDisconnected:
      case eEngineServerStatePauseConnected:
        // Pause is executed deferred with the next playback cycle.
        break;
      case eEngineServerStateDraining:
          FsmUp(&server,eEngineServerRequest,__FILE__,__LINE__,__func__);
          EngineServerHandshake(&server);
          continue;
      case eEngineServerStateBase:
        if (eEngineServerRequest<eEngineServerStateOpen) {
          // When already killed we shutdown the server immediately.
          FsmDown(&server,eEngineServerRequest );
          EngineServerHandshake(&server);
          break;
        }

        // Shutting down the server is executed deferred with the next
        // playback cycle.
        break;
      case eEngineServerStateStart:
        EngineServerHandshake(&server);
        break;
      default:
        DVMESSAGE("unexpectedly received %d",eEngineServerRequest);
        break;
      }

      continue;
    case WAIT_OBJECT_0+eTimerEvent:
    case WAIT_OBJECT_0+eTimer:
      switch (eEngineServer) {
      case eEngineServerStateEot:
        FsmDown(&server,eEngineServerStatePauseConnected );
        server.pAudioClient->lpVtbl->Reset(server.pAudioClient);
        FsmDown(&server,eEngineServerStateBase );
DWRITELN("===============================================================");
DWRITELN("===============================================================");
DWRITELN("===============================================================");
DWRITELN("===============================================================");
DWRITELN("===============================================================");
        EngineServerHandshake(&server);
        break;
      default:
        // We might be in state eEngineServerStateFlush.
        if (eEngineServerStateOpen==eEngineServerRequest) {
          FsmUp(&server,eEngineServerRequest,__FILE__,__LINE__,__func__);
          EngineServerHandshake(&server);
        }

        if (eEngineServerStateOpen<=eEngineServer&&EngineServerRender(&server,eEngineServer,eEngineServerRequest)<0) {
          DMESSAGE("rendering");
          goto end;
        }

        break;
      }

      continue;
    case WAIT_ABANDONED_0+eThreadEvent:
      DVMESSAGE("WAIT_ABANDONED (%s) eThreadEvent",
          FsmStateId(&server));
      continue;
    case WAIT_ABANDONED_0+eTimerEvent:
      DVMESSAGE("WAIT_ABANDONED (%s) eTimerEvent",
          FsmStateId(&server));
      continue;
    case WAIT_ABANDONED_0+eTimer:
      DVMESSAGE("WAIT_ABANDONED (%s) eTimer",FsmStateId(&server));
      continue;
    case WAIT_TIMEOUT:
      DVMESSAGE("WAIT_TIMEOUT (%s)",FsmStateId(&server));
      continue;
    case WAIT_FAILED:
      DVMESSAGE("WAIT_FAILED (%s): %lu",FsmStateId(&server),
          GetLastError());
      continue;
    default:
      DVMESSAGE("unexpectedly received (%s) %d",
          FsmStateId(&server),dwWait);
      continue;
    }
  }
end:
  --server.pClient->fsm.pState;
  EngineServerHandshake(&server);
  PlayerUnlock(pPlayer); // }
  SwitchToThread();
eopen:
  EngineServerDestroy(&server);
ecreate:
  ExitThread(0ul);
  return 0ul;
}

///////////////////////////////////////////////////////////////////////////////
int EngineServerCreate(EngineServer *pServer, EngineClient *pClient)
{
  memset(pServer,0,sizeof *pServer);
  pServer->pClient=pClient;
  pClient->pServer=pServer;
  FsmCreate(pServer,gcaIEngineServerState,eEngineServerStateCount ,&pClient->pPlayer->gtk,
      &pClient->pPlayer->gtk.config.config.monitor.state.server);

  return 0;
}

void EngineServerDestroy(EngineServer *pServer)
{
  pServer->pClient->pServer=NULL;
  FsmDestroy(pServer);
}

///////////////////////////////////////////////////////////////////////////////
int EngineServerNumFramesPadding(EngineServer *pServer,
    UINT32 *puNumFramesPadding)
{
  EEngineServer eEngineServer=FsmState(pServer);
  EngineClient *pClient=pServer->pClient;
  IAudioClient *pAudioClient=pServer->pAudioClient;
  HRESULT hr;

  switch (eEngineServer) {
  case eEngineServerStateFlush:
  case eEngineServerStateKill:
  case eEngineServerStatePauseConnected:
  case eEngineServerStatePauseDisconnected:
    *puNumFramesPadding=0;
    goto success;
  default:
    if (eEngineServer<eEngineServerStateOpen) {
      DVMESSAGE("unexpected state %s",FsmStateId(pServer));
      goto exit;
    }

    break;
  }

  DASSERT(pClient,exit);

  if (pClient->param.bDoubleBuffered)
    *puNumFramesPadding=pClient->param.uNumFramesDevice;
  else {
    hr=pAudioClient->lpVtbl->GetCurrentPadding(pAudioClient,
      puNumFramesPadding    // [out] UINT32 *pNumPaddingFrames
    );

    if (FAILED(hr)) {
#define DCLABEL "IAudioClient::GetCurrentPadding" // [
      switch (hr) {
      DCMESSAGE(DCLABEL,AUDCLNT_E_NOT_INITIALIZED);
      DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_INVALIDATED);
      DCMESSAGE(DCLABEL,AUDCLNT_E_SERVICE_NOT_RUNNING);
      DCMESSAGE(DCLABEL,E_POINTER);
      DDMESSAGE(DCLABEL,hr);
      }
#undef DCLABEL // ]
      goto exit;
    }
  }
success:
  return 0;
exit:
  return -1;
}

///////////////////////////////////////////////////////////////////////////////
int _EngineServerRender(EngineServer *pServer, UINT32 uFrames, int bSilence)
{
  int code=-1;
  EngineClient *pClient=pServer->pClient;
  WORD nBlockAlign=pClient->wfx.target.Format.nBlockAlign;
  DWORD dwSizeRead=nBlockAlign*uFrames;
  GtkClient *pGtk=&pClient->pPlayer->gtk;
  ConfigMonitor *pMonitor=&pGtk->config.config.monitor;
  Ring *pRing=&pClient->ring;
  DWORD dwSize=RingTargetSize(pRing);
  DWORD dwWritten=RingTargetSizeWritten(pRing);
  IAudioRenderClient *pAudioRenderClient=pServer->pAudioRenderClient;
  OptionsDevice *pOptionsDevice=&pClient->options.device;
  DWORD dwFlags=0ul;
  BYTE *pData;
  HRESULT hr;

  if (FsmState(pServer)<eEngineServerStateFlush) {
    DVMESSAGE("unexpected state %s",FsmStateId(pServer));
    goto exit;
  }

  pClient->monitor.xWritten=(double)dwWritten/dwSize;
  GtkClientUpdateWidget(pGtk,&pMonitor->written);

  if (dwSizeRead<=dwWritten) {
    // Everything is fine.
  }
  else if (FsmState(pServer)<eEngineServerStateDraining) {
    if (!bSilence) {
      // We're not already draining and there are not enough frames
      // available. Hence we leave out this cycle. Instead we insert
      // silence and count the event as dropout.
      bSilence=1;
      YASAPI_ADD_UINT32(pClient->stats.uNumFramesDropout,uFrames);
      GtkClientUpdateWidget(pGtk,&pMonitor->engine.num_frames_dropout);
      YASAPI_ADD_DOUBLE(pClient->pPlayer->stats.xMsDropout,
          YASAPI_FRAMES2MS_DOUBLE(uFrames,&pClient->wfx.target));
      GtkClientUpdateWidget(pGtk,&pMonitor->player.ms_dropout);
    }
  }
  else /**/if (!pClient->param.bDoubleBuffered)/**/ {
    // eEngineServerStateDraining && underflow: we've to render the final frames!
    dwSizeRead=dwWritten;
    uFrames=YASAPI_SIZE2FRAMES(dwWritten,&pClient->wfx.target);
  }

  /////////////////////////////////////////////////////////////////////////////
  hr=pAudioRenderClient->lpVtbl->GetBuffer(pAudioRenderClient,
    uFrames,                // [in]  UINT32 NumFramesRequested,
    &pData                  // [out] BYTE   **ppData
  );

  if (FAILED(hr)) {
#define DCLABEL "IAudioRenderClient::GetBuffer" // [
    switch (hr) {
    DVCMESSAGE(DCLABEL ": uNumFrams=%u (%u)",AUDCLNT_E_BUFFER_ERROR,
        uFrames,pClient->param.uNumFramesDevice);
    DVCMESSAGE(DCLABEL ": uNumFrams=%u (%u)",AUDCLNT_E_BUFFER_TOO_LARGE,
        uFrames,pClient->param.uNumFramesDevice);
    DVCMESSAGE(DCLABEL ": uNumFrams=%u (%u)",AUDCLNT_E_BUFFER_SIZE_ERROR,
        uFrames,pClient->param.uNumFramesDevice);
    DCMESSAGE(DCLABEL,AUDCLNT_E_OUT_OF_ORDER);
    DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_BUFFER_OPERATION_PENDING);
    DCMESSAGE(DCLABEL,AUDCLNT_E_SERVICE_NOT_RUNNING);
    DCMESSAGE(DCLABEL,E_POINTER);
    DDMESSAGE(DCLABEL,hr);
#undef DCLABEL // ]
    }

    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (bSilence) {
    ZeroMemory(pData,dwSizeRead);
    dwFlags|=AUDCLNT_BUFFERFLAGS_SILENT;
  }
  else if (dwSizeRead<=dwWritten) {
    // Everything is fine.
    if (RingReadFrames(&pClient->ring,pData,uFrames)<0) {
      DMESSAGE("reading ring");
      goto eread;
    }
  }
  else if (FsmState(pServer)<eEngineServerStateDraining)
    ;
  else {
    // We may go here just when already draining.
    if (dwWritten<dwSizeRead) {
      if (RingReadFrames(&pClient->ring,pData,dwWritten/nBlockAlign)<0) {
        DMESSAGE("reading ring");
        goto eread;
      }

      ZeroMemory(pData+dwWritten,dwSizeRead-dwWritten);
    }
    else {
      ZeroMemory(pData,dwSizeRead);
      dwFlags|=AUDCLNT_BUFFERFLAGS_SILENT;
    }
  }

  code=0;
eread:
  hr=pAudioRenderClient->lpVtbl->ReleaseBuffer(pAudioRenderClient,
    uFrames,                // [in] UINT32 NumFramesWritten,
    dwFlags                 // [in] DWORD  dwFlags
  );

  if (FAILED(hr)) {
#define DCLABEL "IAudioRenderClient::ReleaseBuffer" // [
    switch (hr) {
    DCMESSAGE(DCLABEL,AUDCLNT_E_INVALID_SIZE);
    DCMESSAGE(DCLABEL,AUDCLNT_E_BUFFER_SIZE_ERROR);
    DCMESSAGE(DCLABEL,AUDCLNT_E_OUT_OF_ORDER);
    DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_SERVICE_NOT_RUNNING);
    DCMESSAGE(DCLABEL,E_INVALIDARG);
    DDMESSAGE(DCLABEL,hr);
    }
#undef DCLABEL // ]
    goto exit;
  }

  if (eStrategyPush==pOptionsDevice->eStrategy
      &&!pOptionsDevice->bTimerPeriodic) {
    FsmBacktrack(pServer,eEngineServerStateSetWaitableTimer,__FILE__,__LINE__,__func__);
  }
exit:
  return code;
}

static int EngineServerRender(EngineServer *pServer,
    EEngineServer eEngineServer,
    EEngineServer eEngineServerRequest)
{
  EngineClient *pClient=pServer->pClient;
  GtkClient *pGtk=&pClient->pPlayer->gtk;
  int bSilence=0;
  UINT32 uNumFramesPadding;
  UINT32 uNumFrames;

  /////////////////////////////////////////////////////////////////////////////
  switch (eEngineServerRequest) {
  case eEngineServerStateStart:
    FsmUp(pServer,eEngineServerRequest,__FILE__,__LINE__,__func__);
    EngineServerHandshake(pServer);
    break;
  case eEngineServerStateOpen:
    if (eEngineServerRequest<=FsmState(pServer)) {
      FsmDown(pServer,eEngineServerRequest );
      EngineServerHandshake(pServer);
      break;
    }

    break;
  case eEngineServerStatePauseConnected:
    FsmDown(pServer,eEngineServerRequest );
    EngineServerHandshake(pServer);

    if (pClient->param.bDoubleBuffered)
      break;

    goto success;
  case eEngineServerStateKill:
    FsmDown(pServer,eEngineServerRequest );
    EngineServerHandshake(pServer);
    goto success;
  case eEngineServerStateBase:
  case eEngineServerStatePauseDisconnected:
    FsmDown(pServer,eEngineServerRequest );
    EngineServerHandshake(pServer);
    goto success;
  default:
    DVMESSAGE("unexpected state %s",eEngineServerRequest);
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
  switch (eEngineServer) {
  case eEngineServerStateFlush:
  case eEngineServerStatePauseConnected:
    if (!pClient->param.bDoubleBuffered)
      goto success;

    bSilence=1;
    break;
  case eEngineServerStateKill:
  case eEngineServerStatePauseDisconnected:
    goto success;
  case eEngineServerStateDraining:
    if (0ul==RingTargetSizeWritten(&pClient->ring)) {
      FsmUp(pServer,eEngineServerStateEot,__FILE__,__LINE__,__func__);
      goto success;
    }

DVWRITELN("PRIO: %d",GetThreadPriority(GetCurrentThread()));
    break;
  default:
    if (eEngineServer<eEngineServerStatePauseConnected)
      goto success;

    break;
  }

  if (pClient->param.bDoubleBuffered)
    uNumFrames=pClient->param.uNumFramesDevice;
  else {
    if (!pServer->pAudioClient) {
      DMESSAGE("attempt to access invalidated pAudioClient");
      goto exit;
    }

    if (EngineServerNumFramesPadding(pServer,&uNumFramesPadding)<0) {
      DMESSAGE("getting padding");
      goto exit;
    }

    pClient->monitor.xPadding
        =(double)uNumFramesPadding/pClient->param.uNumFramesDevice;
    GtkClientUpdateWidget(pGtk,&pGtk->config.config.monitor.padding);

    if (0u==uNumFramesPadding)
      uNumFrames=pClient->param.uNumFramesDevice;
    else if (pClient->param.uNumFramesDevice<uNumFramesPadding) {
      // Should never happen, but unfortunately sometimes it does.
      uNumFrames=0u;
    }
    else
      uNumFrames=pClient->param.uNumFramesDevice-uNumFramesPadding;
  }

  if (_EngineServerRender(pServer,uNumFrames,bSilence)<0) {
    DMESSAGE("rendering");
    goto exit;
  }

  /////////////////////////////////////////////////////////////////////////////
success:
  return 0;
exit:
  return -1;
}

///////////////////////////////////////////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////////
static void EngineServerHandshake(EngineServer *pServer)
{
  pServer->thread.eStateRequest=eEngineServerStateStart;
  SetEvent(pServer->pClient->thread.hEvent);
}

static void EngineServerError(EngineServer *pServer)
{
  pServer->thread.eStateRequest=eEngineServerStateError;
  SetEvent(pServer->pClient->thread.hEvent);
}
