/*
 * yasapi_producer_volume.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

///////////////////////////////////////////////////////////////////////////////
#define YASAPI_ID_VOLUME(id) \
  YASAPI_PASTE3(EngineServerVolume,YASAPI_ID,id)
#define YASAPI_ID_IVOLUME_VTBL(id) \
  YASAPI_PASTE3(gcIEngineServerVolume,id,Vtbl)

///////////////////////////////////////////////////////////////////////////////
#define YASAPI_ID Null // [
static int YASAPI_ID_VOLUME(Create)(EngineServer *pServer)
{
  return 0;
}

static void YASAPI_ID_VOLUME(Destroy)(EngineServer *pServer)
{
}

static void YASAPI_ID_VOLUME(Set)(EngineServer *pServer)
{
}

IEngineServerVolumeVtbl YASAPI_ID_IVOLUME_VTBL(YASAPI_ID)={
  .pId="Off",
  .bApplyGain=0,
  .Create=YASAPI_ID_VOLUME(Create),
  .Destroy=YASAPI_ID_VOLUME(Destroy),
  .Set=YASAPI_ID_VOLUME(Set),
};
#undef YASAPI_ID // ]

///////////////////////////////////////////////////////////////////////////////
#define YASAPI_ID Software // [
static int YASAPI_ID_VOLUME(Create)(EngineServer *pServer)
{
  return 0;
}

static void YASAPI_ID_VOLUME(Destroy)(EngineServer *pServer)
{
}

static void YASAPI_ID_VOLUME(Set)(EngineServer *pServer)
{
}

IEngineServerVolumeVtbl YASAPI_ID_IVOLUME_VTBL(YASAPI_ID)={
  .pId="YASAPI/NT",
  .bApplyGain=1,
  .Create=YASAPI_ID_VOLUME(Create),
  .Destroy=YASAPI_ID_VOLUME(Destroy),
  .Set=YASAPI_ID_VOLUME(Set),
};

static IEngineServerVolumeVtbl YASAPI_ID_IVOLUME_VTBL(SoftwareAudioSimple)={
  .pId="YASAPI/NT (Fallback for ISimpleAudioVolume)",
  .bApplyGain=1,
  .Create=YASAPI_ID_VOLUME(Create),
  .Destroy=YASAPI_ID_VOLUME(Destroy),
  .Set=YASAPI_ID_VOLUME(Set),
};

static IEngineServerVolumeVtbl YASAPI_ID_IVOLUME_VTBL(SoftwareAudioEndpoint)={
  .pId="YASAPI/NT (Fallback for IAudioEndpointVolume)",
  .bApplyGain=1,
  .Create=YASAPI_ID_VOLUME(Create),
  .Destroy=YASAPI_ID_VOLUME(Destroy),
  .Set=YASAPI_ID_VOLUME(Set),
};
#undef YASAPI_ID // ]

///////////////////////////////////////////////////////////////////////////////
#define YASAPI_ID SimpleAudio // [
static void YASAPI_ID_VOLUME(Set)(EngineServer *pServer);

static int YASAPI_ID_VOLUME(Create)(EngineServer *pServer)
{
  IAudioClient *pAudioClient=pServer->pAudioClient;
  EngineClient *pClient=pServer->pClient;
  EngineServerVolumeSimple *pSimple=&pClient->param.volume.simple;
  ISimpleAudioVolume **ppSimpleAudioVolume=&pSimple->pSimpleAudioVolume;
  HRESULT hr;

  hr=pAudioClient->lpVtbl->GetService(pAudioClient,
    &IID_ISimpleAudioVolume,        // [in]  REFIID riid,
    (void **)ppSimpleAudioVolume    // [out] void   **ppv
  );

  if (FAILED(hr)) {
#define DCLABEL "IAudioClient::GetService<ISimpleAudioVolume>" // [
    switch (hr) {
    DCMESSAGE(DCLABEL,E_POINTER);
    DCMESSAGE(DCLABEL,E_NOINTERFACE);
    DCMESSAGE(DCLABEL,AUDCLNT_E_NOT_INITIALIZED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_WRONG_ENDPOINT_TYPE);
    DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_SERVICE_NOT_RUNNING);
    DDMESSAGE(DCLABEL,hr);
    }
#undef DCLABEL // ]
    goto eactivate;
  }

  YASAPI_ID_VOLUME(Set)(pServer);

  return 0;
eactivate:
  pClient->param.volume.lpVtbl=&YASAPI_ID_IVOLUME_VTBL(SoftwareAudioSimple);

  return pClient->param.volume.lpVtbl->Create(pServer);
}

static void YASAPI_ID_VOLUME(Destroy)(EngineServer *pServer)
{
  EngineClient *pClient=pServer->pClient;
  EngineServerVolumeSimple *pSimple=&pClient->param.volume.simple;
  ISimpleAudioVolume *pSimpleAudioVolume=pSimple->pSimpleAudioVolume;

  pSimpleAudioVolume->lpVtbl->Release(pSimpleAudioVolume);
}

static void YASAPI_ID_VOLUME(Set)(EngineServer *pServer)
{
  EngineClient *pClient=pServer->pClient;
  EngineServerVolumeSimple *pSimple=&pClient->param.volume.simple;
  ISimpleAudioVolume *pSimpleAudioVolume=pSimple->pSimpleAudioVolume;
  int nVolume=pClient->pPlayer->param.nVolume;
  float fLevel;
  HRESULT hr;

  if (nVolume<0)
    goto exit;

  fLevel=ePlayerVolumeMax<nVolume?1.0:(double)nVolume/ePlayerVolumeMax;

  hr=pSimpleAudioVolume->lpVtbl->SetMasterVolume(pSimpleAudioVolume,
    fLevel,       // [in] float   fLevel,
    NULL          // [in] LPCGUID EventContext
  );

  if (FAILED(hr)) {
#define DCLABEL "SimpleAudioVolume::SetMasterVolume" // [
    switch (hr) {
    DCMESSAGE(DCLABEL,E_INVALIDARG);
    DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_INVALIDATED);
    DCMESSAGE(DCLABEL,AUDCLNT_E_SERVICE_NOT_RUNNING);
    DDMESSAGE(DCLABEL,hr);
    }
#undef DCLABEL // ]
    goto exit;
  }
exit:
  return;
}

IEngineServerVolumeVtbl YASAPI_ID_IVOLUME_VTBL(YASAPI_ID)={
  .pId="I" YASAPI_STRA(YASAPI_ID) "Volume",
  .bApplyGain=0,
  .Create=YASAPI_ID_VOLUME(Create),
  .Destroy=YASAPI_ID_VOLUME(Destroy),
  .Set=YASAPI_ID_VOLUME(Set),
};
#undef YASAPI_ID // ]

///////////////////////////////////////////////////////////////////////////////
#define YASAPI_ID AudioEndpoint // [
static void YASAPI_ID_VOLUME(Set)(EngineServer *pServer);

static int YASAPI_ID_VOLUME(Create)(EngineServer *pServer)
{
  EngineClient *pClient=pServer->pClient;
  IMMDevice *pDevice=pClient->param.pDevice;
  EngineServerVolumeEndpoint *pEndpoint=&pClient->param.volume.endpoint;
  IAudioEndpointVolume **ppAudioEndpointVolume=&pEndpoint->pAudioEndpointVolume;
  DWORD dwHardwareSupportMask;
  HRESULT hr;

  /////////////////////////////////////////////////////////////////////////////
  hr=pDevice->lpVtbl->Activate(pDevice,
    &IID_IAudioEndpointVolume,      // [in]  REFIID      iid,
    CLSCTX_ALL,                     // [in]  DWORD       dwClsCtx,
    NULL,                           // [in]  PROPVARIANT *pActivationParams,
    (void **)ppAudioEndpointVolume  // [out] void        **ppInterface
  );

  if (FAILED(hr)) {
#define DCLABEL "IMMDevice::Activate<IAudioEndpointVolume>" // [
    switch (hr) {
    DCMESSAGE(DCLABEL,E_NOINTERFACE);
    DCMESSAGE(DCLABEL,E_POINTER);
    DCMESSAGE(DCLABEL,E_INVALIDARG);
    DCMESSAGE(DCLABEL,E_OUTOFMEMORY);
    DCMESSAGE(DCLABEL,AUDCLNT_E_DEVICE_INVALIDATED);
    DDMESSAGE(DCLABEL,hr);
    }
#undef DCLABEL // ]
    goto eactivate;
  }

  /////////////////////////////////////////////////////////////////////////////
  hr=(*ppAudioEndpointVolume)->lpVtbl->QueryHardwareSupport(
    *ppAudioEndpointVolume,
    &dwHardwareSupportMask          // [out] DWORD *pdwHardwareSupportMask
  );

  if (FAILED(hr)) {
#define DCLABEL "IAudioEndpointVolume::QueryHardwareSupport" // [
    switch (hr) {
    DCMESSAGE(DCLABEL,E_POINTER);
    DDMESSAGE(DCLABEL,hr);
    }
#undef DCLABEL // ]
    goto ehardware;
  }
  else if (!(ENDPOINT_HARDWARE_SUPPORT_VOLUME&dwHardwareSupportMask))
    goto ehardware;

  /////////////////////////////////////////////////////////////////////////////
  hr=(*ppAudioEndpointVolume)->lpVtbl->GetMasterVolumeLevelScalar(
    *ppAudioEndpointVolume,
    &pEndpoint->fLevel               // [out] float *pfLevel
  );

  if (FAILED(hr)) {
#define DCLABEL "IAudioEndpointVolume::GetMasterVolumeLevelScalart" // [
    switch (hr) {
    DCMESSAGE(DCLABEL,E_POINTER);
    DDMESSAGE(DCLABEL,hr);
    }
#undef DCLABEL // ]
    goto elevel;
  }

  /////////////////////////////////////////////////////////////////////////////
  hr=(*ppAudioEndpointVolume)->lpVtbl->GetVolumeRange(*ppAudioEndpointVolume,
    &pEndpoint->fVolumeMindB,       // float *pflVolumeMindB,
    &pEndpoint->fVolumeMaxdB,       // float *pflVolumeMaxdB,
    &pEndpoint->fVolumeIncrementdB  // float *pflVolumeIncrementdB
  );

  if (FAILED(hr)) {
#define DCLABEL "IAudioEndpointVolume::GetVolumeRange" // [
    switch (hr) {
    DCMESSAGE(DCLABEL,E_POINTER);
    DDMESSAGE(DCLABEL,hr);
    }
#undef DCLABEL // ]
    goto erange;
  }

  YASAPI_ID_VOLUME(Set)(pServer);

  return 0;
//error:
  (*ppAudioEndpointVolume)->lpVtbl->SetMasterVolumeLevelScalar(
    *ppAudioEndpointVolume,
    pEndpoint->fLevel,          // float   fLevel,
    NULL                        // LPCGUID pguidEventContext
  );
erange:
elevel:
ehardware:
  (*ppAudioEndpointVolume)->lpVtbl->Release(*ppAudioEndpointVolume);
eactivate:
  pClient->param.volume.lpVtbl=&YASAPI_ID_IVOLUME_VTBL(SoftwareAudioEndpoint);

  return pClient->param.volume.lpVtbl->Create(pServer);
}

static void YASAPI_ID_VOLUME(Destroy)(EngineServer *pServer)
{
  EngineClient *pClient=pServer->pClient;
  EngineServerVolumeEndpoint *pEndpoint=&pClient->param.volume.endpoint;
  IAudioEndpointVolume *pAudioEndpointVolume=pEndpoint->pAudioEndpointVolume;

  pAudioEndpointVolume->lpVtbl->SetMasterVolumeLevelScalar(
    pAudioEndpointVolume,
    pEndpoint->fLevel,      // float   fLevel,
    NULL                    // LPCGUID pguidEventContext
  );

  pAudioEndpointVolume->lpVtbl->Release(pAudioEndpointVolume);
}

static void YASAPI_ID_VOLUME(Set)(EngineServer *pServer)
{
  EngineClient *pClient=pServer->pClient;
  EngineServerVolumeEndpoint *pEndpoint=&pClient->param.volume.endpoint;
  IAudioEndpointVolume *pAudioEndpointVolume=pEndpoint->pAudioEndpointVolume;
  int nVolume=pServer->pClient->pPlayer->param.nVolume;
  float fLevel;
  HRESULT hr;

  fLevel=nVolume<=0?0.0:nVolume<ePlayerVolumeMax?(double)nVolume/ePlayerVolumeMax:1.0;
DVWRITELN("%f (%d %d)",fLevel,nVolume,ePlayerVolumeMax);

  hr=pAudioEndpointVolume->lpVtbl->SetMasterVolumeLevelScalar(
    pAudioEndpointVolume,
    fLevel,     // [in] float   fLevel,
    NULL        // [in] LPCGUID pguidEventContext
  );

  if (FAILED(hr)) {
    switch (hr) {
    DCMESSAGE("IAudioEndpointVolume::SetMasterVolumeLevelScalar",
        E_INVALIDARG);
    DCMESSAGE("IAudioEndpointVolume::SetMasterVolumeLevelScalar",
        E_OUTOFMEMORY);
    DDMESSAGE("IAudioEndpointVolume::SetMasterVolumeLevelScalar",hr);
    }
  }
}

IEngineServerVolumeVtbl YASAPI_ID_IVOLUME_VTBL(YASAPI_ID)={
  .pId="I" YASAPI_STRA(YASAPI_ID) "Volume",
  .bApplyGain=0,
  .Create=YASAPI_ID_VOLUME(Create),
  .Destroy=YASAPI_ID_VOLUME(Destroy),
  .Set=YASAPI_ID_VOLUME(Set),
};
#undef YASAPI_ID // ]
