/*
 * yasapi_engine.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>

void FsmCreate(void *pFsm, const void *pState, int nStates , GtkClient *pGtkClient,
    ConfigWidget *pWidget)
{
  ((Fsm *)pFsm)->pStateMin=pState;
  ((Fsm *)pFsm)->pStateMax=((Fsm *)pFsm)->pStateMin+nStates+1;
  ((Fsm *)pFsm)->pState=pState;
  ((Fsm *)pFsm)->pGtkClient=pGtkClient;
  ((Fsm *)pFsm)->pWidget=pWidget;
  ((Fsm *)pFsm)->pId="";
DVWRITELN(">>> %p %p <<<",pGtkClient,pWidget);
}

void FsmDestroy(void *pe)
{
  FsmDown(pe,0 );
}

int FsmState(void *pe)
{
  Fsm *pFsm=pe;

  return pe?pFsm->pState-pFsm->pStateMin:0;
}

const char *FsmStateIdEx(void *pe, int nState)
{
  Fsm *pFsm=pe;
  IFsmState *pStateMin=pFsm->pStateMin;
  IFsmState *pState=pFsm->pStateMin+nState;

  return pState<=pStateMin?"eFsmStart":(pState-1)->pId;
}

const char *FsmStateId(void *pe)
{
  Fsm *pFsm=pe;
  IFsmState *pState=pFsm->pState;
  IFsmState *pStateMin=pFsm->pStateMin;

  return FsmStateIdEx(pFsm,pState-pStateMin);
}

int FsmBacktrack(void *pe, int nState, const char *file, int line, const char *func)
{
  Fsm *pFsm=pe;
  int nStateCur=FsmState(pFsm);

  FsmDown(pFsm,nState-1 );

  if (FsmUp(pFsm,nStateCur,file,line,func)<0)
    goto exit;

  return 0;
exit:
  return -1;
}

static void FsmUpdateWidget(Fsm *pFsm)
{
  if (0<FsmState(pFsm->pGtkClient)&&pFsm->pWidget) {
    pFsm->pId=FsmStateId(pFsm);
DVWRITELN(">>> %s <<<",pFsm->pId);
    GtkClientUpdateWidget(pFsm->pGtkClient,pFsm->pWidget);
  }
}

int FsmUp(void *pe, int nState, const char *file, int line, const char *func)
{
  Fsm *pFsm=pe;
  IFsmState *pStateTarget=pFsm->pStateMin+nState;

  if (pFsm->pStateMax<=pStateTarget) {
    DMESSAGEEX(file,line,func,"state overflow");
    goto exit;
  }

  if (pFsm->pState<pStateTarget) {
    do {
      DVWRITELNEX(file,line,func,"entering %s ...",pFsm->pState->pId);

      if (pFsm->pState->Enter(pFsm)<0) {
        DVMESSAGEEX(file,line,fun,"entering %s",pFsm->pState->pId);
        goto exit;
      }

    FsmUpdateWidget(pFsm);
    } while (pFsm->pState<pStateTarget);
  }

  return nState;
exit:
  return -1;
}

void FsmDown(void *pe, int nState )
{
  Fsm *pFsm=pe;
  IFsmState *pStateTarget=pFsm->pStateMin+nState;

  if (pStateTarget<pFsm->pState) {
    do {
      DVWRITELNEX(file,line,func,"leaving %s ...",(pFsm->pState-1)->pId);
      (pFsm->pState-1)->Leave(pFsm);
      FsmUpdateWidget(pFsm);
    } while (pStateTarget<pFsm->pState);
  }
}
