/*
 * yasapi_gtk_server.c
 *
 * Yet Another (WA)SAPI (YASAPI) New Technology (NT) Output (out_yasapi-nt)
 * Copyright (C) 2015-2019  Peter Belkner (pbelkner@snafu.de)
 *
 * out_yasapi-nt is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * out_yasapi-nt is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with out_yasapi-nt.  If not, see <https://www.gnu.org/licenses/>.
 */
#include <yasapi.h>
#include <winamp/wa_ipc.h>

///////////////////////////////////////////////////////////////////////////////
static int GtkServerCreate(GtkServer *pServer, GtkClient *pClient);
static void GtkServerDestroy(GtkServer *pServer);

DWORD WINAPI GtkServerThreadProc(LPVOID lParameter)
{
  GtkServer server;

  if (GtkServerCreate(&server,lParameter)<0)
    goto ecreate;

  if (FsmUp(&server,eGtkServerStateBase,__FILE__,__LINE__,__func__)<0)
    goto eopen;

  ++server.pClient->fsm.pState;
  SetEvent(server.pClient->thread.hEvent);
  SwitchToThread();

  if (FsmUp(&server,eGtkServerStateReady,__FILE__,__LINE__,__func__)<0)
    goto exit;

  if (server.pClient->pDialog)
    GtkServerDialogShowSync (server.pClient->pDialog);

  gtk_main();
exit:
  FsmDown(&server,eGtkServerStateStart );
  --server.pClient->fsm.pState;
  SetEvent(server.pClient->thread.hEvent);
  SwitchToThread();
eopen:
  GtkServerDestroy(&server);
ecreate:
  ExitThread(0ul);
  return 0ul;
}

///////////////////////////////////////////////////////////////////////////////
static int GtkServerCreate(GtkServer *pServer, GtkClient *pClient)
{
  memset(pServer,0,sizeof *pServer);
  FsmCreate(pServer,gcaIGtkServerState,eGtkServerStateCount ,NULL,NULL);

  pServer->pClient=pClient;
  pClient->pServer=pServer;

  return 0;
}

static void GtkServerDestroy(GtkServer *pServer)
{
  pServer->pClient->pServer=NULL;
  pServer->pClient=NULL;
  FsmDestroy(pServer);
}

///////////////////////////////////////////////////////////////////////////////

void GtkServerDialogWrite(Dialog *pDialog)
{
  Player *pPlayer=pDialog->pGtk->pPlayer;

  if (pDialog->window&&gtk_widget_get_visible (pDialog->window)) {
DWRITELN("|||||||||||||||||||||||||||||||||||||||||||||||||||||");
    pDialog->lpVtbl->PropertiesWrite(pDialog,pPlayer);
DWRITELN("|||||||||||||||||||||||||||||||||||||||||||||||||||||");
  }
}

void gtk_dialog_options_save (GtkWidget *widget, gpointer data)
{
  GtkServerDialogWrite(data);
}

gboolean window_delete_event (GtkWidget *window, GdkEvent *event,
    gpointer user_data)
{
  Dialog *pDialog=user_data;

  GtkServerDialogWrite(pDialog);
  gtk_widget_hide (window);

  return TRUE;
}

///////////////////////////////////////////////////////////////////////////////
//#define YASAPI_GTK_WINDOW_PRESENT
#if ! defined (YASAPI_GTK_WINDOW_PRESENT)  // [
static void GtkServerDialogForeground (Dialog *pDialog)
{
  HWND hWnd=GDK_WINDOW_HWND(gtk_widget_get_window (pDialog->window));

  SetForegroundWindow(hWnd);
}
#endif // ]

static void GtkServerDialogActivate (Dialog *pDialog)
{
  HWND hWnd=GDK_WINDOW_HWND(gtk_widget_get_window (pDialog->window));
  HWND hWndParent=pDialog->hWndParent;

  pDialog->hWndParent=NULL;
  SetWindowLongPtrW(hWnd,GWL_HWNDPARENT,(LONG_PTR)hWndParent);
#if defined (YASAPI_GTK_WINDOW_PRESENT)  // [
  gtk_window_present (GTK_WINDOW(pDialog->window));
#else // ] [
  GtkServerDialogForeground (pDialog);
#endif // ]
}

void GtkServerDialogMove (Dialog *pDialog)
{
  if (!pDialog->bMoved) {
    gtk_widget_show_all (pDialog->window);
    pDialog->lpVtbl->Finalize(pDialog);

    pDialog->bMoved=1;
  }
}

void GtkServerDialogShowSync (Dialog *pDialog)
{
  if (!pDialog||!pDialog->window)
    goto exit;

  if (!pDialog->bMoved)
    GtkServerDialogMove (pDialog);
  else
    gtk_widget_show(pDialog->window);

  GtkServerDialogActivate (pDialog);
  pDialog->pGtk->pDialog=NULL;
exit:
  return;
}

gboolean GtkServerDialogShowAsync (gpointer user_data)
{
  GtkServerDialogShowSync (user_data);

  return FALSE;
}

///////////////////////////////////////////////////////////////////////////////
static void GtkServerPresentSync (GtkServer *pServer)
{
  GtkClient *pClient=pServer->pClient;
  Dialog *pDialogMax=pClient->aDialog+eGtkNumDialogs;
  Dialog *pDialog;

  for (pDialog=pClient->aDialog;pDialog<pDialogMax;++pDialog) {
    if (pDialog->window&&gtk_widget_get_visible(pDialog->window))
#if defined (YASAPI_GTK_WINDOW_PRESENT)  // [
      gtk_window_present (GTK_WINDOW(pDialog->window));
#else // ] [
      GtkServerDialogForeground (pDialog);
#endif // ]
  }
}

gboolean GtkServerPresentAsync (gpointer user_data)
{
  GtkServerPresentSync (user_data);

  return FALSE;
}
